import os
from pathlib import Path
import sys

__all__ = [
    'get_app_data_parent',
]


def get_app_data_parent() -> Path:
    """Get the parent directory for the app's data directory."""
    home = Path.home()
    if os.name == 'nt':  # Windows
        return home / 'AppData' / 'Local'
    elif os.name == 'posix':
        if sys.platform == 'darwin':  # MacOS
            return home / 'Library' / 'Application Support'
        else:  # Linux
            return Path(os.getenv('XDG_DATA_HOME', home / '.local' / 'share'))
    else:
        return home



from __future__ import annotations
from dataclasses import dataclass

from reling.utils.prompts import enter_to_continue, Prompt, PromptOption
from reling.tts import TTSVoiceClient
from reling.types import Reader, Speed
from reling.utils.console import clear_current_line
from reling.utils.types import ensure_not_none

__all__ = [
    'output',
    'SentenceData',
]

PROMPT_TITLE = 'Play'
NORMAL_SPEED = 'normal speed'
SLOWLY = 'slowly'
REPLAY = 'replay'


@dataclass
class SentenceData:
    text: str
    print_prefix: str = ''
    reader: Reader | None = None
    reader_id: str | None = None

    @staticmethod
    def from_tts(
            text: str,
            client: TTSVoiceClient | None,
            *,
            print_prefix: str = '',
            reader_id: str | None = None,
    ) -> SentenceData:
        return SentenceData(
            text=text,
            print_prefix=print_prefix,
            reader=client.get_reader(text) if client else None,
            reader_id=reader_id,
        )


@dataclass
class ReaderWithSpeed:
    reader: Reader
    speed: Speed


@dataclass
class ReaderWithId:
    reader: Reader
    id: str


def add_single_sentence_options(prompt: Prompt[ReaderWithSpeed], reader: Reader) -> None:
    """Attach the options for a single sentence to the prompt: '[n]ormal speed | [s]lowly'."""
    prompt.add_option(PromptOption(
        description=NORMAL_SPEED,
        action=ReaderWithSpeed(reader, Speed.NORMAL),
    ))
    prompt.add_option(PromptOption(
        description=SLOWLY,
        action=ReaderWithSpeed(reader, Speed.SLOW),
    ))


def add_multi_sentence_options(prompt: Prompt[ReaderWithSpeed], readers: list[ReaderWithId]) -> None:
    """Attach the options for multiple sentences to the prompt: '[i]mproved | [is] | [o]riginal | [os]'."""
    for reader in readers:
        prompt.add_option(PromptOption(
            description=reader.id,
            action=ReaderWithSpeed(reader.reader, Speed.NORMAL),
            modifiers={
                SLOWLY: ReaderWithSpeed(reader.reader, Speed.SLOW),
            },
        ))


def construct_prompt(
        sentences_with_readers: list[SentenceData],
        current: ReaderWithSpeed | None,
        multi_sentence: bool,
) -> Prompt:
    """
    Construct a prompt for the user to choose the next sentence to read and the speed of the reading.
    :raises ValueError: If reader_id is not provided for a sentence with a reader in a multi-sentence output.
    """
    prompt = Prompt(PROMPT_TITLE)
    if multi_sentence:
        add_multi_sentence_options(prompt, [
            ReaderWithId(
                reader=ensure_not_none(sentence.reader),
                id=ensure_not_none(sentence.reader_id),
            )
            for sentence in sentences_with_readers
        ])
    else:
        add_single_sentence_options(prompt, sentences_with_readers[0].reader)
    if current:
        prompt.add_option(PromptOption(
            description=REPLAY,
            action=current,
        ))
    return prompt


def output(*sentences: SentenceData) -> None:
    """
    Output the sentences, reading them if a reader is provided.
    If multiple readers are provided, the user can choose which sentence to read next.
    The user can also choose the speed of the reading.

    :raises ValueError: If reader_id is not provided for a sentence with a reader in a multi-sentence output.
    """
    for sentence in sentences:
        print(sentence.print_prefix + sentence.text)
    multi_sentence = len(sentences) > 1
    if sentences_with_readers := [sentence for sentence in sentences if sentence.reader]:
        current = ReaderWithSpeed(sentences_with_readers[0].reader, Speed.NORMAL) if len(sentences) == 1 else None
        while True:
            if current:
                current.reader(current.speed)
                clear_current_line()  # Otherwise the input made during the reading will get displayed twice
            current = construct_prompt(sentences_with_readers, current, multi_sentence).prompt()
            if not current:
                break
    elif multi_sentence:
        enter_to_continue()



from dataclasses import dataclass
from datetime import timedelta
from functools import partial
from typing import Iterable

from reling.helpers.output import output, SentenceData
from reling.helpers.wave import play
from reling.tts import TTSVoiceClient
from reling.types import DialogueExchangeData, Input
from reling.utils.scores import format_average_score
from reling.utils.time import format_time_delta
from reling.utils.transformers import get_numbering_prefix
from .scoring import MAX_SCORE
from .types import ExchangeWithTranslation, ScoreWithSuggestion, SentenceWithTranslation

__all__ = [
    'present_dialogue_results',
    'present_text_results',
]

NA = 'N/A'


@dataclass
class TitleData:
    text: str
    should_number: bool
    tts: TTSVoiceClient | None


def present_results(
        titles: Iterable[list[TitleData]],
        provided_translations: Iterable[Input],
        original_translations: Iterable[str],
        results: Iterable[ScoreWithSuggestion],
        duration: timedelta,
        target_tts: TTSVoiceClient | None,
) -> None:
    """Present the results of scoring translations."""
    scores: list[int] = []
    for index, (title_items, provided_translation, original_translation, result) in enumerate(zip(
            titles,
            provided_translations,
            original_translations,
            results,
    )):
        scores.append(result.score)
        for title in title_items:
            output(SentenceData.from_tts(
                title.text,
                title.tts,
                print_prefix=get_numbering_prefix(index) if title.should_number else '',
            ))
        print(f'Your score: {result.score}/{MAX_SCORE}')
        output(
            SentenceData(
                text=provided_translation.text.strip() or NA,
                print_prefix='Provided: ',
                reader=partial(play, provided_translation.audio) if provided_translation.audio and target_tts else None,
                reader_id='provided',
            ),
            SentenceData.from_tts(
                text=result.suggestion or NA,
                client=target_tts if result.suggestion else None,
                print_prefix='Improved: ',
                reader_id='improved',
            ),
            SentenceData.from_tts(
                text=original_translation,
                client=target_tts,
                print_prefix='Original: ',
                reader_id='original',
            ),
        )
        print()
    print(f'Average score: {format_average_score(scores)}')
    print(f'Exam duration: {format_time_delta(duration)}')


def present_text_results(
        sentences: Iterable[SentenceWithTranslation],
        original_translations: Iterable[str],
        results: Iterable[ScoreWithSuggestion],
        duration: timedelta,
        source_tts: TTSVoiceClient | None,
        target_tts: TTSVoiceClient | None,
) -> None:
    """Present the results of scoring text translations."""
    present_results(
        titles=([
            TitleData(
                text=sentence.sentence,
                should_number=True,
                tts=source_tts,
            ),
        ] for sentence in sentences),
        provided_translations=(sentence.translation for sentence in sentences),
        original_translations=original_translations,
        results=results,
        duration=duration,
        target_tts=target_tts,
    )


def present_dialogue_results(
        exchanges: Iterable[ExchangeWithTranslation],
        original_translations: Iterable[DialogueExchangeData],
        results: Iterable[ScoreWithSuggestion],
        duration: timedelta,
        source_user_tts: TTSVoiceClient | None,
        target_speaker_tts: TTSVoiceClient | None,
        target_user_tts: TTSVoiceClient | None,
) -> None:
    """Present the results of scoring dialogue translations."""
    present_results(
        titles=([
            TitleData(
                text=original_translation.speaker,
                should_number=False,
                tts=target_speaker_tts,
            ),
            TitleData(
                text=exchange.exchange.user,
                should_number=True,
                tts=source_user_tts,
            ),
        ] for original_translation, exchange in zip(
            original_translations,
            exchanges,
        )),
        provided_translations=(exchange.user_translation for exchange in exchanges),
        original_translations=(exchange.user for exchange in original_translations),
        results=results,
        duration=duration,
        target_tts=target_user_tts,
    )



from pathlib import Path
from tempfile import TemporaryDirectory

from tqdm import tqdm

from reling.app.app import app
from reling.app.exceptions import AlgorithmException
from reling.app.translation import get_dialogue_exchanges, get_text_sentences
from reling.app.types import (
    API_KEY,
    ASR_MODEL,
    CONTENT_ARG,
    LANGUAGE_OPT,
    LANGUAGE_OPT_FROM,
    LISTEN_OPT,
    MODEL,
    READ_LANGUAGE_OPT,
    TTS_MODEL,
)
from reling.asr import ASRClient
from reling.db.models import Dialogue, Language, Text
from reling.gpt import GPTClient
from reling.helpers.pyaudio import ensure_pyaudio
from reling.helpers.voices import pick_voices
from reling.tts import TTSClient
from reling.utils.functions import promisify
from reling.utils.time import now
from reling.utils.typer import typer_raise
from .input import collect_dialogue_translations, collect_text_translations
from .presentation import present_dialogue_results, present_text_results
from .scoring import score_dialogue_translations, score_text_translations
from .storage import save_dialogue_exam, save_text_exam
from .types import ExchangeWithTranslation, SentenceWithTranslation

__all__ = [
    'exam',
]


@app.command()
def exam(
        api_key: API_KEY,
        model: MODEL,
        tts_model: TTS_MODEL,
        asr_model: ASR_MODEL,
        content: CONTENT_ARG,
        from_: LANGUAGE_OPT_FROM = None,
        to: LANGUAGE_OPT = None,
        read: READ_LANGUAGE_OPT = None,
        listen: LISTEN_OPT = False,
) -> None:
    """
    Test the user's ability to translate content from one language to another.
    If only one language is specified, the content's original language is assumed for the unspecified direction.
    """
    if read or listen:
        ensure_pyaudio()
    if from_ is None and to is None:
        typer_raise('You must specify at least one language.')
    from_ = from_ or content.language
    to = to or content.language
    if from_ == to:
        typer_raise('The source and target languages are the same.')

    read = read or []
    for language in read:
        if language not in [from_, to]:
            typer_raise(f'Cannot read in {language.name} as it is not the source or target language.')

    (perform_text_exam if isinstance(content, Text) else perform_dialogue_exam)(
        GPTClient(api_key=api_key.get(), model=model.get()),
        content,
        from_,
        to,
        source_tts=TTSClient(api_key=api_key.get(), model=tts_model.get()) if from_ in read else None,
        target_tts=TTSClient(api_key=api_key.get(), model=tts_model.get()) if to in read else None,
        asr=ASRClient(api_key=api_key.get(), model=asr_model.get()) if listen else None,
    )


def perform_text_exam(
        gpt: GPTClient,
        text: Text,
        source_language: Language,
        target_language: Language,
        source_tts: TTSClient | None,
        target_tts: TTSClient | None,
        asr: ASRClient | None,
) -> None:
    """
    Translate the text as needed, collect user translations, score them, save and present the results to the user,
    optionally reading the source and/or target language out loud.
    """
    with TemporaryDirectory() as file_storage:
        source_voice, target_voice = pick_voices(None, None)
        voice_source_tts = source_tts.with_voice(source_voice) if source_tts else None
        voice_target_tts = target_tts.with_voice(target_voice) if target_tts else None

        sentences = get_text_sentences(promisify(gpt), text, source_language)
        original_translations = get_text_sentences(promisify(gpt), text, target_language)

        started_at = now()
        translated = list(collect_text_translations(
            sentences=sentences,
            target_language=target_language,
            source_tts=voice_source_tts,
            asr=asr,
            storage=Path(file_storage),
        ))
        finished_at = now()

        try:
            results = list(tqdm(
                score_text_translations(
                    gpt=gpt,
                    sentences=translated,
                    original_translations=original_translations,
                    source_language=source_language,
                    target_language=target_language,
                ),
                desc='Scoring translations',
                total=len(translated),
            ))
        except AlgorithmException as e:
            typer_raise(e.msg)

        save_text_exam(
            text=text,
            source_language=source_language,
            target_language=target_language,
            read_source=source_tts is not None,
            read_target=target_tts is not None,
            listened=asr is not None,
            started_at=started_at,
            finished_at=finished_at,
            sentences=translated,
            results=results,
        )

        present_text_results(
            sentences=translated,
            original_translations=original_translations,
            results=results,
            duration=finished_at - started_at,
            source_tts=voice_source_tts,
            target_tts=voice_target_tts,
        )


def perform_dialogue_exam(
        gpt: GPTClient,
        dialogue: Dialogue,
        source_language: Language,
        target_language: Language,
        source_tts: TTSClient | None,
        target_tts: TTSClient | None,
        asr: ASRClient | None,
) -> None:
    """
    Translate the dialogue as needed, collect user translations, score them, save and present the results to the user,
    optionally reading the source and/or target language out loud.
    """
    with TemporaryDirectory() as file_storage:
        speaker_voice, user_voice = pick_voices(dialogue.speaker_gender, dialogue.user_gender)
        source_user_tts = source_tts.with_voice(user_voice) if source_tts else None
        target_user_tts = target_tts.with_voice(user_voice) if target_tts else None
        target_speaker_tts = target_tts.with_voice(speaker_voice) if target_tts else None

        exchanges = get_dialogue_exchanges(promisify(gpt), dialogue, source_language)
        original_translations = get_dialogue_exchanges(promisify(gpt), dialogue, target_language)

        started_at = now()
        translated = list(collect_dialogue_translations(
            exchanges=exchanges,
            original_translations=original_translations,
            target_language=target_language,
            source_user_tts=source_user_tts,
            target_speaker_tts=target_speaker_tts,
            asr=asr,
            storage=Path(file_storage),
        ))
        finished_at = now()

        try:
            results = list(tqdm(
                score_dialogue_translations(
                    gpt=gpt,
                    exchanges=translated,
                    original_translations=original_translations,
                    source_language=source_language,
                    target_language=target_language,
                ),
                desc='Scoring translations',
                total=len(translated),
            ))
        except AlgorithmException as e:
            typer_raise(e.msg)

        save_dialogue_exam(
            dialogue=dialogue,
            source_language=source_language,
            target_language=target_language,
            read_source=source_tts is not None,
            read_target=target_tts is not None,
            listened=asr is not None,
            started_at=started_at,
            finished_at=finished_at,
            exchanges=translated,
            results=results,
        )

        present_dialogue_results(
            exchanges=translated,
            original_translations=original_translations,
            results=results,
            duration=finished_at - started_at,
            source_user_tts=source_user_tts,
            target_speaker_tts=target_speaker_tts,
            target_user_tts=target_user_tts,
        )



try:
    import readline  # noqa: F401 (https://stackoverflow.com/a/14796424/430083)
except ImportError:
    pass  # Windows

import typer

__all__ = [
    'app',
]

app = typer.Typer(pretty_exceptions_enable=False)
